# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================
# matplotlib is loaded in GPy initialization and if a backend('Agg') is not set, loading fails in mac osx
import matplotlib
matplotlib.use('Agg')

# import repurposer base classes
from .repurposer import Repurposer  # noqa: F401
from .meta_model_repurposer import MetaModelRepurposer  # noqa: F401
from .neural_network_repurposer import NeuralNetworkRepurposer  # noqa: F401

# import repurposers
from .svm_repurposer import SvmRepurposer  # noqa: F401
from .lr_repurposer import LrRepurposer  # noqa: F401
from .bnn_repurposer import BnnRepurposer  # noqa: F401
from .gp_repurposer import GpRepurposer  # noqa: F401
from .neural_network_fine_tune_repurposer import NeuralNetworkFineTuneRepurposer  # noqa: F401
from .neural_network_random_freeze_repurposer import NeuralNetworkRandomFreezeRepurposer  # noqa: F401

# import load function
from .load_repurposer import load  # noqa: F401

# import version
from .__version__ import __version__  # noqa: F401
